import fsMod from "node:fs";
import { createHash } from "node:crypto";
import { addRollupInput } from "../add-rollup-input.js";
import {} from "../internal.js";
import { generateContentEntryFile, generateLookupMap } from "../../../content/vite-plugin-content-virtual-mod.js";
import { joinPaths } from "../../path.js";
import { fileURLToPath, pathToFileURL } from "node:url";
import { CONTENT_RENDER_FLAG } from "../../../content/consts.js";
import { copyFiles } from "../static-build.js";
import pLimit from "p-limit";
const CONTENT_CACHE_DIR = "./content/";
const CONTENT_MANIFEST_FILE = "./manifest.json";
const CONTENT_ENTRY_CACHE_DIR = "./entries/";
const CONTENT_MANIFEST_VERSION = 0;
const virtualEmptyModuleId = `virtual:empty-content`;
const resolvedVirtualEmptyModuleId = `\0${virtualEmptyModuleId}`;
function vitePluginContent(opts, lookupMap, _internals) {
  const { cacheDir } = opts.settings.config;
  const contentCacheDir = new URL(CONTENT_CACHE_DIR, cacheDir);
  const contentManifestFile = new URL(CONTENT_MANIFEST_FILE, contentCacheDir);
  const contentEntryCacheDir = new URL(CONTENT_ENTRY_CACHE_DIR, contentCacheDir);
  let oldManifest = { version: -1, entries: [] };
  let newManifest = { version: -1, entries: [] };
  let entries;
  if (fsMod.existsSync(contentManifestFile)) {
    try {
      const data = fsMod.readFileSync(contentManifestFile, { encoding: "utf8" });
      oldManifest = JSON.parse(data);
    } catch {
    }
  }
  return {
    name: "@astro/plugin-build-content",
    async options(options) {
      let newOptions = Object.assign({}, options);
      if (opts.settings.config.output === "static") {
        newManifest = await generateContentManifest(opts, lookupMap);
        await fsMod.promises.mkdir(new URL("./", contentManifestFile), { recursive: true });
        await fsMod.promises.writeFile(contentManifestFile, JSON.stringify(newManifest), { encoding: "utf8" });
        entries = getEntriesFromManifests(oldManifest, newManifest);
        console.log(`Building ${entries.buildFromSource.length} changed files...`);
        for (const { type, entry } of entries.buildFromSource) {
          const fileURL = joinPaths(opts.settings.config.root.toString(), entry);
          const input = fileURLToPath(fileURL);
          const inputs = [`${input}?${collectionTypeToFlag(type)}`];
          if (type === "content") {
            inputs.push(`${input}?${CONTENT_RENDER_FLAG}`);
          }
          newOptions = addRollupInput(newOptions, inputs);
        }
        if (entries.buildFromSource.length === 0) {
          newOptions = addRollupInput(newOptions, [virtualEmptyModuleId]);
        }
      }
      return newOptions;
    },
    resolveId(id) {
      if (id === virtualEmptyModuleId) {
        return resolvedVirtualEmptyModuleId;
      }
    },
    load(id) {
      if (id === resolvedVirtualEmptyModuleId) {
        return {
          code: `// intentionally left empty!
export default {}`
        };
      }
    },
    async generateBundle(options, bundle) {
      const content = await generateContentEntryFile({ settings: opts.settings, fs: fsMod, lookupMap });
      this.emitFile({
        type: "prebuilt-chunk",
        code: content,
        fileName: ".entry.mjs"
      });
      Object.keys(bundle).forEach((key) => {
        const mod = bundle[key];
        if (mod.type === "asset")
          return;
        if (mod.facadeModuleId === resolvedVirtualEmptyModuleId) {
          delete bundle[key];
        }
      });
    },
    async writeBundle(options) {
      const dist = new URL(pathToFileURL(options.dir).toString() + "/");
      const cache = contentEntryCacheDir;
      const cacheTmp = new URL("./.tmp/", cache);
      const cacheExists = fsMod.existsSync(cache);
      fsMod.mkdirSync(cache, { recursive: true });
      await fsMod.promises.mkdir(cacheTmp, { recursive: true });
      await copyFiles(dist, cacheTmp, true);
      if (cacheExists) {
        await copyFiles(contentEntryCacheDir, dist, false);
      }
      await copyFiles(cacheTmp, contentEntryCacheDir);
      await fsMod.promises.rm(cacheTmp, { recursive: true, force: true });
    }
  };
}
function getEntriesFromManifests(oldManifest, newManifest) {
  const { version: oldVersion, entries: oldEntries } = oldManifest;
  const { version: newVersion, entries: newEntries } = newManifest;
  let entries = { restoreFromCache: [], buildFromSource: [] };
  const newEntryMap = new Map(newEntries);
  if (oldVersion !== newVersion || oldEntries.length === 0) {
    entries.buildFromSource = Array.from(newEntryMap.keys());
    return entries;
  }
  const oldEntryHashMap = new Map(oldEntries.map(([key, hash]) => [hash, key]));
  for (const [entry, hash] of newEntryMap) {
    if (oldEntryHashMap.has(hash)) {
      entries.restoreFromCache.push(entry);
    } else {
      entries.buildFromSource.push(entry);
    }
  }
  return entries;
}
async function generateContentManifest(opts, lookupMap) {
  let manifest = { version: CONTENT_MANIFEST_VERSION, entries: [] };
  const limit = pLimit(10);
  const promises = [];
  for (const [collection, { type, entries }] of Object.entries(lookupMap)) {
    for (const entry of Object.values(entries)) {
      const key = { collection, type, entry };
      const fileURL = new URL(joinPaths(opts.settings.config.root.toString(), entry));
      promises.push(limit(async () => {
        const data = await fsMod.promises.readFile(fileURL, { encoding: "utf8" });
        manifest.entries.push([key, checksum(data)]);
      }));
    }
  }
  await Promise.all(promises);
  return manifest;
}
function checksum(data) {
  return createHash("sha1").update(data).digest("base64");
}
function collectionTypeToFlag(type) {
  const name = type[0].toUpperCase() + type.slice(1);
  return `astro${name}CollectionEntry`;
}
function pluginContent(opts, internals) {
  return {
    targets: ["content"],
    hooks: {
      async "build:before"() {
        const lookupMap = await generateLookupMap({ settings: opts.settings, fs: fsMod });
        return {
          vitePlugin: vitePluginContent(opts, lookupMap, internals)
        };
      }
    }
  };
}
export {
  pluginContent
};
